"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeterministicFiniteAutomaton = void 0;
const FiniteAutomaton_1 = require("./FiniteAutomaton");
class DeterministicFiniteAutomaton extends FiniteAutomaton_1.FiniteAutomaton {
    constructor(testLogic, automaton, automatonId) {
        super(testLogic, automaton, 'deterministic', automatonId);
    }
    #generateMergedDfaData(state, newStates, newTransitions, newFinalStates, mergeOperation, inputAutomaton, isComposite, separator) {
        const currentAutomatonFinalStates = new Set(this.automaton.final_states);
        const inputAutomatonFinalStates = new Set(inputAutomaton ? inputAutomaton.automaton.final_states : []);
        this.automaton.states.forEach((currentDfaState) => {
            const newState = isComposite ? `${currentDfaState}${separator}${state}` : currentDfaState;
            if (isComposite) {
                newStates.push(newState);
                newTransitions[newState] = {};
                this.automaton.alphabets.forEach((automatonAlphabet) => {
                    newTransitions[newState][automatonAlphabet] = [
                        this.automaton.transitions[currentDfaState][automatonAlphabet] +
                            separator +
                            (inputAutomaton
                                ? inputAutomaton.automaton.transitions[state][automatonAlphabet]
                                : ''),
                    ];
                });
                if (mergeOperation === 'or' &&
                    ((inputAutomaton ? inputAutomatonFinalStates.has(state) : true) ||
                        currentAutomatonFinalStates.has(currentDfaState))) {
                    newFinalStates.add(newState);
                }
                else if (mergeOperation === 'and' &&
                    (inputAutomaton ? inputAutomatonFinalStates.has(state) : true) &&
                    currentAutomatonFinalStates.has(currentDfaState)) {
                    newFinalStates.add(newState);
                }
                else if (mergeOperation === 'not' && !currentAutomatonFinalStates.has(currentDfaState)) {
                    newFinalStates.add(newState);
                }
            }
            else {
                if (mergeOperation === 'or' &&
                    ((inputAutomaton ? inputAutomatonFinalStates.has(newState) : true) ||
                        currentAutomatonFinalStates.has(newState))) {
                    newFinalStates.add(newState);
                }
                else if (mergeOperation === 'and' &&
                    (inputAutomaton ? inputAutomatonFinalStates.has(newState) : true) &&
                    currentAutomatonFinalStates.has(newState)) {
                    newFinalStates.add(newState);
                }
                else if (mergeOperation === 'not' && !currentAutomatonFinalStates.has(newState)) {
                    newFinalStates.add(newState);
                }
            }
        });
    }
    #merge(finiteAutomaton, mergeOperation, mergedDfaOptions) {
        const { separator = '.', label, description } = mergedDfaOptions ?? {};
        const newStates = [];
        const newTransitions = {};
        const isComposite = Boolean(finiteAutomaton &&
            (finiteAutomaton ? finiteAutomaton.getAutomatonId() : '') !== this.getAutomatonId());
        const newDfaId = isComposite && finiteAutomaton
            ? this.getAutomatonId() + separator + finiteAutomaton.getAutomatonId()
            : this.getAutomatonId();
        const newStartState = isComposite && finiteAutomaton
            ? this.automaton.start_state + separator + finiteAutomaton.automaton.start_state
            : this.automaton.start_state;
        const newFinalStates = new Set();
        if (finiteAutomaton) {
            finiteAutomaton.automaton.states.forEach((state) => {
                this.#generateMergedDfaData(state, newStates, newTransitions, newFinalStates, mergeOperation, finiteAutomaton, isComposite, separator);
            });
        }
        else {
            this.#generateMergedDfaData('', newStates, newTransitions, newFinalStates, mergeOperation, finiteAutomaton, isComposite, separator);
        }
        return new DeterministicFiniteAutomaton((inputString, automatonTestResult) => {
            if (mergeOperation === 'or') {
                return (finiteAutomaton.testLogic(inputString, automatonTestResult) ||
                    this.testLogic(inputString, automatonTestResult));
            }
            else if (mergeOperation === 'and') {
                return (finiteAutomaton.testLogic(inputString, automatonTestResult) &&
                    this.testLogic(inputString, automatonTestResult));
            }
            else {
                return !this.testLogic(inputString, automatonTestResult);
            }
        }, {
            final_states: Array.from(newFinalStates),
            label: label ??
                mergeOperation +
                    '(' +
                    `${this.automaton.label}${mergeOperation !== 'not' ? ', ' : ''}${finiteAutomaton ? finiteAutomaton.automaton.label : ''}` +
                    ')',
            description: description ??
                mergeOperation.toUpperCase() +
                    '(' +
                    `${this.automaton.description}${mergeOperation !== 'not' ? ', ' : ''}${finiteAutomaton ? finiteAutomaton.automaton.description : ''}` +
                    ')',
            start_state: isComposite ? newStartState : this.automaton.start_state,
            states: isComposite ? newStates : JSON.parse(JSON.stringify(this.automaton.states)),
            transitions: (isComposite
                ? newTransitions
                : JSON.parse(JSON.stringify(this.automaton.transitions))),
            alphabets: this.automaton.alphabets,
            epsilon_transitions: null,
        }, isComposite ? newDfaId : this.getAutomatonId());
    }
    AND(dfaModule, mergedDfaOptions) {
        return this.#merge(dfaModule, 'and', mergedDfaOptions);
    }
    NOT(mergedDfaOptions) {
        return this.#merge(undefined, 'not', mergedDfaOptions);
    }
    OR(dfaModule, mergedDfaOptions) {
        return this.#merge(dfaModule, 'or', mergedDfaOptions);
    }
}
exports.DeterministicFiniteAutomaton = DeterministicFiniteAutomaton;
