"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FiniteAutomataTest = void 0;
const cli_progress_1 = __importDefault(require("cli-progress"));
const colors_1 = __importDefault(require("colors"));
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const utils_1 = require("../utils");
const GenerateString_1 = require("./GenerateString");
class FiniteAutomataTest {
    #cliProgressBar;
    #logsPath;
    constructor(logsPath) {
        this.#cliProgressBar = new cli_progress_1.default.SingleBar({
            barsize: 20,
            stopOnComplete: true,
            format: colors_1.default.green('{bar}') + ' {percentage}% {value}/{total} Chunks',
        }, cli_progress_1.default.Presets.shades_classic);
        this.#logsPath = logsPath;
        if (!fs_1.default.existsSync(logsPath)) {
            fs_1.default.mkdirSync(logsPath);
        }
    }
    #createFileWriteStreams(automatonLabel, outputFiles) {
        const writeStreamsRecord = {
            caseWriteStream: outputFiles.case
                ? fs_1.default.createWriteStream(path_1.default.resolve(this.#logsPath, `${automatonLabel}.case.txt`))
                : null,
            aggregateWriteStream: outputFiles.aggregate
                ? fs_1.default.createWriteStream(path_1.default.resolve(this.#logsPath, `${automatonLabel}.aggregate.txt`))
                : null,
            incorrectWriteStream: outputFiles.incorrect
                ? fs_1.default.createWriteStream(path_1.default.resolve(this.#logsPath, `${automatonLabel}.incorrect.txt`))
                : null,
            correctWriteStream: outputFiles.correct
                ? fs_1.default.createWriteStream(path_1.default.resolve(this.#logsPath, `${automatonLabel}.correct.txt`))
                : null,
            inputWriteStream: outputFiles.input
                ? fs_1.default.createWriteStream(path_1.default.resolve(this.#logsPath, `${automatonLabel}.input.txt`))
                : null,
            acceptedWriteStream: outputFiles.accepted
                ? fs_1.default.createWriteStream(path_1.default.resolve(this.#logsPath, `${automatonLabel}.accepted.txt`))
                : null,
            rejectedWriteStream: outputFiles.rejected
                ? fs_1.default.createWriteStream(path_1.default.resolve(this.#logsPath, `${automatonLabel}.rejected.txt`))
                : null,
        };
        return {
            record: writeStreamsRecord,
            endStreams() {
                Object.values(writeStreamsRecord).forEach((writeStream) => {
                    if (writeStream) {
                        writeStream.end();
                    }
                });
            },
        };
    }
    #testAutomata(finiteAutomaton, finiteAutomatonTestInfo, writeStreams, inputStrings) {
        const { caseWriteStream, correctWriteStream, acceptedWriteStream, rejectedWriteStream, incorrectWriteStream, inputWriteStream, } = writeStreams;
        for (let i = 0; i < inputStrings.length; i++) {
            const inputString = inputStrings[i].replace('\r', '').replace('\n', '');
            if (inputString.length !== 0) {
                const { automatonTestResult } = finiteAutomaton.generateGraphFromString(inputString);
                const logicTestResult = finiteAutomaton.testLogic(inputString, automatonTestResult);
                const isWrong = automatonTestResult !== logicTestResult;
                const testResultString = automatonTestResult.toString().toUpperCase()[0] +
                    ' ' +
                    logicTestResult.toString().toUpperCase()[0] +
                    ' ' +
                    inputString +
                    ' ' +
                    '\n';
                if (!automatonTestResult) {
                    rejectedWriteStream && rejectedWriteStream.write(inputString + '\n');
                }
                else {
                    acceptedWriteStream && acceptedWriteStream.write(inputString + '\n');
                }
                if (!isWrong) {
                    if (automatonTestResult === false && logicTestResult === false) {
                        finiteAutomatonTestInfo.trueNegatives += 1;
                    }
                    else {
                        finiteAutomatonTestInfo.truePositives += 1;
                    }
                    correctWriteStream && correctWriteStream.write(testResultString);
                }
                else {
                    if (automatonTestResult && !logicTestResult) {
                        finiteAutomatonTestInfo.falsePositives += 1;
                    }
                    else {
                        finiteAutomatonTestInfo.falseNegatives += 1;
                    }
                    incorrectWriteStream && incorrectWriteStream.write(testResultString);
                }
                inputWriteStream && inputWriteStream.write(inputString + '\n');
                const { withoutColors } = (0, utils_1.generateCaseMessage)(isWrong, inputString, automatonTestResult, logicTestResult);
                caseWriteStream && caseWriteStream.write(withoutColors + '\n');
                this.#cliProgressBar.increment(1);
            }
        }
    }
    #postTest(finiteAutomaton, finiteAutomatonTestInfo, finiteAutomatonWriteStreams) {
        const { record: { aggregateWriteStream }, endStreams, } = finiteAutomatonWriteStreams;
        const { withoutColors, withColors } = (0, utils_1.generateAggregateMessage)(finiteAutomaton.label, finiteAutomaton.description, finiteAutomatonTestInfo);
        console.log('\n' + withColors);
        aggregateWriteStream && aggregateWriteStream.write(withoutColors);
        endStreams();
    }
    async test(configs) {
        const finiteAutomatonTestInfos = configs.map(() => ({
            falsePositives: 0,
            falseNegatives: 0,
            truePositives: 0,
            trueNegatives: 0,
        }));
        for (let index = 0; index < configs.length; index++) {
            const config = configs[index];
            const finiteAutomatonTestInfo = finiteAutomatonTestInfos[index];
            const { automaton, options } = config;
            const writeStreams = this.#createFileWriteStreams(automaton.automaton.label, {
                aggregate: options.outputFiles?.aggregate ?? true,
                case: options.outputFiles?.case ?? true,
                correct: options.outputFiles?.correct ?? true,
                incorrect: options.outputFiles?.incorrect ?? true,
                input: options.outputFiles?.input ?? true,
                accepted: options.outputFiles?.accepted ?? true,
                rejected: options.outputFiles?.rejected ?? true,
            });
            if (options.type === 'file') {
                const readStream = fs_1.default.createReadStream(options.filePath);
                const fileLines = await (0, utils_1.countFileLines)(options.filePath);
                this.#cliProgressBar.start(fileLines, 0, {
                    speed: 'N/A',
                });
                for await (const chunks of readStream) {
                    const inputStrings = chunks.toString().split('\n');
                    this.#testAutomata(automaton, finiteAutomatonTestInfo, writeStreams.record, inputStrings);
                }
            }
            else {
                let generatedStrings = [];
                if (options.type === 'generate') {
                    if (options.random) {
                        generatedStrings = GenerateString_1.GenerateString.generateRandomUnique(options.random.total, options.random.minLength, options.random.maxLength);
                    }
                    else if (options.range) {
                        generatedStrings = GenerateString_1.GenerateString.generateAllCombosWithinLength(automaton.automaton.alphabets, options.range.maxLength);
                    }
                }
                else {
                    generatedStrings = options.inputs;
                }
                this.#cliProgressBar.start(generatedStrings.length, 0, {
                    speed: 'N/A',
                });
                this.#testAutomata(automaton, finiteAutomatonTestInfo, writeStreams.record, generatedStrings);
            }
            this.#postTest(automaton.automaton, finiteAutomatonTestInfo, writeStreams);
        }
    }
}
exports.FiniteAutomataTest = FiniteAutomataTest;
